% function [w,allLik,allW,allGradSum,it,didConverge] = ...
%    buildLRmodel(data,target,penalty,silent,initialW,weights,negfactor)
%
% Build a logistic regression classifier. 
%
% ASSUMES target \in [0 1]
% 
% penalty term refers to the lambda constant in the term of
% the log likelihood, -lambda*w*w' (i.e. zero-mean gaussian
% prior on the weights, with std. dev w).
%
% NOTE: no penalty on the intercept...
%
% initialW is the intialization, if not provided, starts at zero
%
% if weights are not provided, they are all set to 1.
% weights should be of the number of training points in dd
% and reflect how much each training point should be used during training
%
% if silent=0, then fminunc will print out information at each
% iteration
%
% negfactor is the extra factor applied to negative examples
% (class label 0) in the objective function
% if negfactor > 1, false positives are penalized ?more?
% if negfactor < 1, false positives are penalized ?less?
%
% data should be a vector of dim: [numData numFeat]
%
% Author: Jennifer Listgarten, 2005.

function [w,allLik,allW,allGradSum,it,didConverge] = ...
    nonsymLR(dd,target,penalty,silent,initialW,weights,negfactor)

%METHOD = 'gradDescent';
METHOD = 'fminunc';

% maxIt=1e5;
maxIt=100; % changed by David
convg=1e-5;

uniqueLab=unique(target);
if ~isempty(setdiff(union(uniqueLab,[0,1]),[1,0]))
    error('Labels/targets should be 0/1 only');
end

[numRow,numFeat]=size(dd);

%% first add the bias column to the data
biasD  = [dd, ones(numRow,1)];

if ~exist('initialW') | isempty(initialW)
    initialW = zeros(1,numFeat+1);
end

if ~exist('weights') | isempty(weights)
    weights = ones(numRow,1);
end

if ~exist('negfactor') | isempty(negfactor)
    negfactor = 1;
end

if strcmp(METHOD,'fminunc')
    if silent
        disp = 'off';
        %disp='final';
    else
        disp='iter';
    end    
      
    HessUpdate='bfgs';
    %HessUpdate='dfp';
    %HessUpdate='steepdesc';
    
    options = optimset('GradObj', 'on', 'DerivativeCheck','off',...
        'TolFun',convg,'Display',disp,'TolX',convg,...
        'LargeScale','off','Diagnostics', 'off',...
        'MaxIter',maxIt,'MaxFunEvals',realmax,...
        'Hessian','off','HessUpdate',HessUpdate);  
    
    %% if data has a strange scale, try to make it more normal..
    meanAbs = mean(abs(biasD(:)));
    biasDscale = biasD./meanAbs;
    
    [w, fval, exitflag,output,grad]=...
        fminunc(@nonsymLR_obj_grad,initialW,options,...
                biasDscale,weights,numRow,...
                penalty,target,negfactor);         
%     [w, fval, exitflag,output,grad]=...
%         fminunc(@getLR_obj_grad,initialW,options,...
%                 biasDscale,weights,numRow,...
%                 penalty,target);         
    
     it = output.iterations;
     allLik = -fval;
     allGradSum = abs(sum(grad));
     allW = [];
     output.algorithm;
     didConverge = exitflag>0;
     
elseif (strcmp(METHOD,'gradDescent'))
    momentum=0;
    allW = zeros(maxIt,numFeat+1);
    allLik = zeros(1,maxIt);
    allPred = zeros(numRow,maxIt);
    allGradSum = zeros(1,maxIt);
    it=0;
    keepGoing=1;
    while (keepGoing)
        it=it+1;
        %% this retursn the negative because we use it 
        %% with fminunc
        [tmpLik,deltaW] = ...
            getLR_obj_grad(w,biasD,weights,numFeat,numRow,...
            penalty,target);            
        tmpLik = -tmpLik;
        deltaW = -deltaW;
        w = w + stepSize*deltaW;
        allW(it,:)=w;
        allGradSum(it) = sum(abs(deltaW));       
        allLik(it) = tmpLik;

        %% momentum for step size
        if (momentum & it>1)
            if (diff(allLik([it,it-1]))<0)
                stepSize=stepSize/momentum;
            else
                stepSize=stepSize*momentum;
            end
        end

        if (mod(it,10)==1)
            display(['Iteration ' num2str(it)]);
        end

        if (allGradSum<convg)
            display('Converged');
            didConverge=1;
            keepGoing=0;
        elseif (it>maxIt)
            display('No more iterations allowed');
            didConverge=1;
            keepGoing=0;
        end
    end  %% end iterating

end

display(['Performed ' num2str(it) ' iterations']);

if (0)%(~silent)
    figure, plot(allLik,'+-');
    title(['Penalty=' num2str(penalty) ', iter=' num2str(it)]);
    xlabel('Iteration');
    ylabel('Log likelihood');
    
    if (length(w)<4)  %% can visualize then
        %%% DATA %%%%%%%%%%%%%%%%%%%%%%%%%%
        ls=getLineSpecs(0); figure,
        for cl=1:2
            ind{cl}=(target==cl-1);
            plot(dd(ind{cl},1),dd(ind{cl},2),ls{cl});
            hold on;
        end

        %% find out which points are misclassified
        %% at thresh = 0.5;
        thresh=0.5;
        repW = repmat(w,[numRow,1]);
        z = sum(biasD.*repW,2);
        zLogit = logistic(z);
        pred = zLogit>thresh;
        badInd = abs(target-pred)>0;

        plot(dd(badInd,1),dd(badInd,2),'r*');

        %% line is y=-w1/w2*x -b/w2
        xrg=-4:16;
        plot(xrg,xrg.*(-w(1)/w(2)) - w(3)/w(2),'-');
    end
    
    %figure, plot(allW);
    %legend({'1','2','3'});
end

%% might want to set very small weights to be zero

return







        %% and the predictions if this were the model
        %     pred = zLogit>thresh;
        %     allPred(:,it)=pred;
        %     if (0)%(mod(it,2)==1)
        %         figure, plot(pred);
        %         yaxis([0,1]);
        %         title(['Iter ' num2str(it)]);
        %     end

