function [err, grad_alpha, grad_beta, grad_gamma, grad_eta] = ...
    cdf_error_and_grad(X,Y,alpha,beta,gamma,eta,method)
%[err, grad_alpha, grad_beta, grad_gamma, grad_eta] = ...
%    cdf_error_and_grad(X,Y,alpha,beta,gamma,eta,method)
%
% Required Arguments:
%   X,Y = cell arrays of state and observation sequences
%   alpha = (D*D*J)
%   beta = (D*D*K)
%   gamma = (L*J)
%   eta = (M*K)
%
% Optional Argument:
%   method = either 'sampling' (the default), or 'mean-field'
%
% Return Values:
%   err = a measure of reconstruction error, given the parameters
%   grad_{alpha,beta,gamma,eta} = gradient of the likelihood with
%      respect to indicated parameters.  The resulting matrices are 
%      the same sizes as {alpha,beta,gamma,eta}.

% Author: David Ross, Oct. 2005
% $Id: cdf_error_and_grad.m,v 1.3 2006/11/07 20:24:13 dross Exp $

%-----------------------------------------------------------
% Check the input
% @TODO possibly comment these checks out for speed.
%-----------------------------------------------------------

error(nargchk(6,7,nargin));

if ~iscell(Y)
    error('Y must be a cell array of observation sequences');
end

mean_field = 0;
if exist('method')
    switch lower(method)
    case 'mean-field', mean_field = 1;
    case 'sampling', mean_field = 0;
    otherwise error(sprintf('method %s is unrecognized', method));
    end
end

D = size(alpha,1);
J = size(alpha,3);
K = size(beta,3);
L = size(gamma,1);
M = size(eta,1);
N = length(Y);

if size(alpha,2) ~= D
    error('the size of alpha must be D*D*J');
end

if any(size(beta) ~= [D D K])
    error('the size of beta must be D*D*K');
end

if size(gamma,2) ~= J
    error(sprintf('gamma must have J=%d columns',J));
end

if size(eta,2) ~= K
    error(sprintf('eta must have K=%d columns',K));
end


%-----------------------------------------------------------
% Constants defining behaviour
%-----------------------------------------------------------

ERR_SAMPLES = 10; % number of samples of U,V to take when computing
                  % the error (not used if mean_field)
CD_steps = 2; % number of steps to do in the negative phase of CD
SAMPLES = 100; % number of samples to take in the negative phase


%-----------------------------------------------------------
% Compute the error
%-----------------------------------------------------------

err = 0;
for nn = 1:N
    [pU,pV] = prob_UV_given_XY(X{nn},Y{nn},alpha,beta,gamma,eta);
    if mean_field
        [mu,sigma,mu_pair,sigma_pair] = ...
            prob_X_given_UVY(Y{nn}, pU, pV, alpha, beta);
        err = err + sse(X{nn} - mu);
    else
        for ii = 1:ERR_SAMPLES
            U = sample_u(pU);
            V = pV > rand(size(pV));
            mu = prob_X_given_UVY(Y{nn}, U, V, alpha, beta);
            err = err + sse(X{nn} - mu)/SAMPLES;
        end
    end
end


%-----------------------------------------------------------
% Compute the gradient
%-----------------------------------------------------------

grad_alpha = zeros(D,D,J);
grad_beta = zeros(D,D,K);
grad_gamma = zeros(L,J);
grad_eta = zeros(M,K);


for nn = 1:N
    %% Positive Phase
    %% Compute the expected values of F,G,R, and S, given X and Y
    [EF,EG,ER,ES] = feature_expect_given_X(...
        X{nn}, Y{nn}, alpha, beta, gamma, eta);

    grad_alpha = grad_alpha + EF;
    grad_beta = grad_beta + EG;
    grad_gamma = grad_gamma + ER;
    % grad_eta = grad_eta + ES; % @FULL_ETA
    grad_eta = grad_eta + [diag(ES)'; ES(end,:)]; % @SPARSE_ETA

    
    %% Negative Phase
    for ii = 1:SAMPLES
        % to CD_steps repetitions of "sample U,V given X,Y" and
        % "sample X given U,V,Y"
        X_samp = X{nn};
        for cd = 1:CD_steps
            [pU,pV] = prob_UV_given_XY(X_samp,Y{nn},alpha,beta,gamma,eta);
            if mean_field
                U_samp = pU;
                V_samp = pV;
            else % sampling
                % U_samp = pU > rand(size(pU));
                U_samp = sample_u(pU);
                V_samp = pV > rand(size(pV));
            end
            if cd < CD_steps
                %NOTE: if version 1, we don't need the final X_samp step
                X_samp = sample_x_unigauss(Y{nn}, U_samp, V_samp, alpha, beta);
            end
        end
        
        % Version 1, given a sample of U,V take exact expectation of 
        % features values over X|U,V,Y.
        [EF,EG,ER,ES] = feature_expect_given_UV(...
            U_samp, V_samp, Y{nn}, alpha, beta, gamma, eta);
        
        % Version 2, given a sample of X, take exact expectation of 
        % features over U,V|X,Y.
        %[EF,EG,ER,ES] = feature_expect_given_X(X_samp, Y{nn}, ...
        %    alpha, beta, gamma, eta);
        
        % Version 3, get a sample of X,U,V
        %[EF,EG,ER,ES] = feature_expect_given_XUV(...
        %    X_samp, Y{nn}, U_samp, V_samp, alpha, beta, gamma, eta);
        
        % put it all together 
        if max(abs(EF(:))/100) >1e6
            %disp('found a HUGE gradient');
            %keyboard; % @DEBUG check for huge gradients
        end
        
        grad_alpha = grad_alpha - EF./SAMPLES;
        grad_beta = grad_beta - EG./SAMPLES;
        grad_gamma = grad_gamma - ER./SAMPLES;
        %grad_eta = grad_eta - ES./SAMPLES; % @FULL_ETA
        grad_eta = grad_eta - [diag(ES)'; ES(end,:)]./SAMPLES; % @SPARSE_ETA
    end
end

% @TODO should the grad's be divided by N here?

