function varargout = cdf_infer(Y,alpha,beta,gamma,eta,method,gt)
% [mu,sigma,pU,pV] = cdf_infer(Y,alpha,beta,gamma,eta)
% [mu,sigma,mu_pair,sigma_pair,pU,pV] = cdf_infer(Y,alpha,beta,gamma,eta)
% infer_unigauss(..., 'mean-field')
% infer_unigauss(..., 'sampling')
%
% Required arguments:
%   Y = observation matrix (?*T)
%   alpha, beta, gamma, eta = feature weights
%
% Optional Argument:
%   'sampling' = use sampling for inference (default)
%   'mean-field' = do mean-field inference
%
% Return Values:
%   mu = the marginal means (D*T)
%   sigma = the marginal covariances (D*D*T)
%   mu_pair = the pairwise-marginal means (2D*(T-1))
%   sigma_pair = the pairwise-marginal covariances (2D*2D*(T-1))
%   pU = probability that U indicators are on (J*T)
%   pV =  probability that V indicators are on (K*T)

% Author: David Ross, Oct. 2005
% $Id: cdf_infer.m,v 1.1.1.1 2006/06/01 15:32:22 dross Exp $

%-----------------------------------------------------------
% Check the input.
%-----------------------------------------------------------

error(nargchk(5,7,nargin));
pairwise_marginals = (nargout > 4);

if iscell(Y)
    error('function only works for one Y at a time');
end

mean_field = 0;
if exist('method')
    switch lower(method)
    case 'mean-field', mean_field = 1;
    case 'sampling', mean_field = 0;
    otherwise, error(sprintf('method %s is unrecognized', method));
    end
end

J = size(alpha,3);
K = size(beta,3);

% @TODO add more checks of the input


%-----------------------------------------------------------
% Do filp-flop inference.
%-----------------------------------------------------------

ITERATIONS = 100; %@TODO how many iterations?

% % pick initial values for U and V
[pU,pV] = prob_UV_given_Y(Y,gamma,eta);

%% pick dummy initial values for U and V (old version, doesn't work)
% pU = ones(J,T)*0.3; pU(end,:) = 1;
% pV = ones(K,T)*0.5; pV(end,:) = 1;
if mean_field 
    U = pU;
    V = pV;
else
    % U = rand(size(pU) <= pU;
    U = sample_u(pU);    
    V = rand(size(pV)) <= pV;
end


%% Iteratively update X and U,V
for ii = 1:ITERATIONS
    %% Compute X
    if mean_field
        X = prob_X_given_UVY(Y,U,V,alpha,beta);
    else
        X = sample_x_unigauss(Y,U,V,alpha,beta);
    end
    
    if exist('gt','var') % @DEBUG
        err = sqrt(sum((gt - X([1 4],:)).^2,1));
        disp(sum(err > 5))
    end

    %% Compute U,V
    [pU,pV] = prob_UV_given_XY(X,Y,alpha,beta,gamma,eta);
    if mean_field 
        U = pU;
        V = pV;
    else
        % U = rand(size(pU)) <= pU;
        U = sample_u(pU);
        V = rand(size(pV)) <= pV;
    end
end

if pairwise_marginals
    [mu,sigma,mu_pair,sigma_pair] = prob_X_given_UVY(...
        Y, U, V, alpha, beta);
    varargout = {mu, sigma, mu_pair, sigma_pair, pU, pV};    
else
    [mu,sigma] = prob_X_given_UVY(Y,U,V,alpha,beta);
    varargout = {mu, sigma, pU, pV};
end
