function [alpha,beta,gamma,eta] = cdf_train(X,Y,alpha,beta,gamma,eta)
% [alpha,beta,gamma,eta] = cdf_train(X,Y)
% [alpha,beta,gamma,eta] = cdf_train(X,Y,alpha,beta,gamma,eta)
% Perform contrastive divergence learning of the UniGauss model.
% Use the alternative gradient descent, without an adaptive rate.
%
% Required arguments:
%   X = cell array of label sequences
%   Y = cell array of observation sequences
%
% Optional arguments:
%   alpha, beta, gamma, eta = initial values of weights
%     If omitted or equal to [], then these weights will be initialized
%     randomly.
%
% Return values:
%   alpha, beta, gamma, eta = learned weights

% Author: David Ross, Oct. 2005
% $Id: cdf_train.m,v 1.2 2006/10/04 22:41:34 dross Exp $

%-----------------------------------------------------------
% Check arguments & initialize
%-----------------------------------------------------------

error(nargchk(2,6,nargin));
if ~iscell(X) | ~iscell(Y) | any(any(size(X)~=size(Y)))
    error('X and Y must be cell arrays with the same size');
end

D = size(X{1},1);
J = size(T_matrices,3); % number of dynamics features
K = size(Y{1}(1).nu,2); % new version of Y
% K = compute_g_predictions;
% L = r_features;
% M = s_features;
L = numel(Y{1}(1).rr);
M = numel(Y{1}(1).ss);
N = length(X);


%% Initial guesses of the parameters
small = 1e-2;
if ~exist('alpha','var') | isempty(alpha)
    alpha = repmat(eye(D),[1 1 J])*small;
    alpha = alpha .* rand(size(alpha));
end
if ~exist('beta','var') | isempty(beta)
    beta = repmat(eye(D),[1 1 K])*small;
    beta = beta .* rand(size(beta));
end
if ~exist('gamma','var') | isempty(gamma)
    gamma = randn(L,J)*small;
    % gamma(end,:) = 2;
end
if ~exist('eta','var') | isempty(eta)
% eta = randn(M,K)*small; % @FULL_ETA
    eta = randn(2,K)*small;  % @SPARSE_ETA
    % eta(end,:) = 2;
end


%-----------------------------------------------------------
% Do gradient descent
%-----------------------------------------------------------

ACCEPT_BADSTEPS = 1;
BADSTEP_DECREASE = 4; % @TODO set in an intelligent way
MIN_WEIGHT = 1e-8; % the smallest possible value for the weights
MAX_ITERS = 400;
MAX_BADSTEPS = 10;
MOMENTUM = 0.8;

step_a = 1e-6; % @TODO set in an intelligent way
step_b = 1e-6; % @TODO set in an intelligent way
step_g = 1e-2; % @TODO set in an intelligent way
step_e = 1e-2; % @TODO set in an intelligent way

% compute the gradient
method = 'sampling';
[err, grad_alpha, grad_beta, grad_gamma, grad_eta] = ...
    cdf_error_and_grad(X,Y,alpha,beta,gamma,eta,method);

delta_a = grad_alpha * step_a;
delta_b = grad_beta * step_b;
delta_g = grad_gamma * step_g;
delta_e = grad_eta * step_e;

disp(sprintf('first CD iteration, error %g', err));

num_badsteps = 0;

while length(err) <= MAX_ITERS & num_badsteps < MAX_BADSTEPS
    new_alpha = alpha + delta_a .* repmat(eye(D),[1 1 J]); % @DIAGONAL
    new_beta = beta + delta_b .* repmat(eye(D),[1 1 K]); % @DIAGONAL
    new_gamma = gamma + delta_g;
    new_eta = eta + delta_e;
    
    % make sure alpha and beta stay positive
    % (There should be no restriction on gamma and eta)
    new_alpha = max(new_alpha, repmat(eye(D),[1 1 J])*MIN_WEIGHT);
    new_beta = max(new_beta, repmat(eye(D),[1 1 K])*MIN_WEIGHT);
    
    %     % compute the gradient
    %     if length(err) < 5
    %         method = 'mean-field';
    %     else
    %         method = 'sampling';
    %     end
    [e, grad_alpha, grad_beta, grad_gamma, grad_eta] = ...
        cdf_error_and_grad(X, Y, ...
        new_alpha, new_beta, new_gamma, new_eta, method);

    if e < err(end)
        msg = '--> better';
        num_badsteps = 0;
        if e < min(err)
            save temp.mat alpha beta gamma eta;
            disp('saving (because it got better)');
        end
    else
        msg = '--> worse';
        num_badsteps = num_badsteps+1;
    end

    if ACCEPT_BADSTEPS | e < err(end)
        alpha = new_alpha;
        beta = new_beta;
        gamma = new_gamma;
        eta = new_eta;

        delta_a = grad_alpha * step_a + delta_a * MOMENTUM;
        delta_b = grad_beta * step_b + delta_b * MOMENTUM;
        delta_g = grad_gamma * step_g + delta_g * MOMENTUM;
        delta_e = grad_eta * step_e + delta_e * MOMENTUM;
        
        err(end+1) = e;
    else
        delta_a = delta_a / BADSTEP_DECREASE;
        delta_b = delta_b / BADSTEP_DECREASE;
        delta_g = delta_g / BADSTEP_DECREASE;
        delta_e = delta_e / BADSTEP_DECREASE;
    end
    
    if (length(err)> 2) & (err(end) > err(end-1)*5)
        disp('things just got a lot worse');
        keyboard; % @DEBUG error got >5 times worse
    end

    %disp(sprintf('done CD iteration, mean-field=%d, error %g %s', ...
    %    mean_field, e, msg));
    disp(sprintf('done CD iteration, %s, error %g %s', method, e, msg));
end    

figure;
plot(log(err));
title('cdf_train log error', 'Interpreter', 'none');
xlabel('gradient descent iterations');

disp('Done training, returning the final weights');
if err(end) < min(err)
    disp('(The weights giving the least error are stored in temp.mat');
end
