function Y = compute_features(models, varargin)
% Y = compute_features(models, path1, path2, ...)
% 
% Given a number of video sequences, compute the observation and side-
% information (aka switch potential) feature predictions for each sequence,
% using a trained set of observation models.
%
% Required Arguments:
%    models = Should be the result of 'train_features.m'.
%    path1 = A directory path containing a sequence of images (with
%    filenames in sorted order), to be read by 'load_images.m'.
%
% Return Values:
%    Y = a cell array, one element for each path, each containing a struct
%        array, one element per image in the sequence.  For each image
%        the values computed are: 'nu', a 6*J vector that is the state
%        prediction made by each of the J observation features;
%        'rr' a vector of side-information/switch potentials for the
%        dynamics features, padding with a constant 1 at the end;
%        'ss' a similar vector of side information for the observation
%        features.
%
% The following code will give you a 6*T*K matrix of the state predictions
% made for each timestep of sequence nn by each of the observation
% features: 
%    nu = cat(3,Y{nn}.nu);
%    nu = permute(nu, [1 3 2]);
%
% Similarly, the side-information for sequence nn can be obtained by
%    rr = [Y{nn}.rr];
%    ss = [Y{nn}.ss];
%
% Why have an array of structs instead of a struct of arrays?  Using the
% former allows the sequence to be sliced cleanly, e.g. Y{1}(1:500), while
% the latter does not.

% Author: David Ross, 2006.
% $Id: compute_features.m,v 1.3 2006/07/18 16:25:01 dross Exp $

%----------------------------------------------------------
% Check the input
%----------------------------------------------------------

error(nargchk(2,Inf,nargin));

%----------------------------------------------------------
%----------------------------------------------------------

[HEIGHT,WIDTH] = size(models.pca_mean);
paths = varargin;

N = length(paths);
Y = cell(1,N);

%----------------------------------------------------------
%----------------------------------------------------------

template = double(models.template);
K_pca = size(models.pca_basis,2);
pca_basis_imsize = reshape(models.pca_basis,[HEIGHT WIDTH K_pca]);

% # of observation features, which is equal to the number of templates
% +1 for background subtraction
% +1 for PCA
K = size(models.template,3) + 2;

filt = fspecial('gaussian',11);

for pp = 1:N
    %% load the data
    data_path = varargin{pp};
    data = load_images(data_path);
    T = size(data,3);
    
    % initialize nu and s
    nu = zeros(VALUE_OF_D,K);
    s = ones(K+1,1); % the last feature is an always-on 1

    for tt = 1:T
        % current image
        im = double(data(:,:,tt));

        for jj = 1:size(template,3)
            % apply each of the templates
            output = compare_template(im, template(:,:,jj));
            %figure; imagesc(output);
            [junk, ind] = sort(output(:));
            [rr,cc] = ind2sub(size(im), ind(1));
            nu(1,jj) = cc;
            nu(4,jj) = rr;
            s(jj) = ind(1);
        end

        % background subtraction
        jj = jj+1;
        start = max(1,tt-2);
        stop = min(T,tt+2);
        bg_sub = im - double(mean(data(:,:,start:stop),3));
        bg_sub = imfilter(abs(bg_sub), filt);
        %figure; imagesc(bg_sub);
        [junk, ind] = sort(-bg_sub(:));
        [rr,cc] = ind2sub(size(im), ind(1));
        nu(1,jj) = cc;
        nu(4,jj) = rr;
        s(jj) = ind(1);


        % apply PCA
        jj = jj + 1;
        output = compare_basis2(im, pca_basis_imsize, models.pca_mean);
        [junk, ind] = sort(output(:));
        [rr,cc] = ind2sub(size(im), ind(1));
        nu(1,jj) = cc;
        nu(4,jj) = rr;
        s(jj) = ind(1);

        window = double(im(1:HEIGHT,1:WIDTH));
        window = window - models.pca_mean;
        window = window(:);
        err = sum((window - models.pca_basis*models.pca_basis'*window).^2)
        [i,j] = find(abs(output - err) < 1e-8);
        if i ~= ceil(HEIGHT/2) | j ~= ceil(WIDTH/2)
            warning('compare_basis2 is wrong');
        end

        % display the points
        imagesc(im);
        colormap('gray'); hold on; set(gcf,'doublebuffer','on');
        % scatter(points(1,tt),points(2,tt),'b.');
        scatter(nu(1,1:end-2), nu(4,1:end-2),'rx');
        scatter(nu(1,end-1), nu(4,end-1),'mx');
        scatter(nu(1,end), nu(4,end),'gx');
        text(10,10,num2str(tt),'Color','r');
        hold off;
        drawnow;

        % rescale s
        s = s / 1e5;
        s(end) = 1;
        
        % save the result
        Y{pp}(tt).ss = s;
        Y{pp}(tt).rr = 1;
        Y{pp}(tt).nu = nu;
    end

    % remember: we only have T-1 rr features, since they look at pairs
    % of states, thus Y{pp}(end).rr shouldn't be anything.
    Y{pp}(end).rr = [];

    disp(['done sequence ' data_path]);
end
