function [retval,UUt] = loghmm(A,B,op)
% [marginals] = hmm(A,B)
% [mode] = hmm(A,B,-1)
% [sample] = hmm(A,B,1)
%
% Belief propagation in an undirected Markov chain.
%
% Input Arguments:
%   A = pairwise potential matrix.  A(j,i) is the (time invariant)
%       potential of seeing i and time t and j at time t+1.
%       For an HMM, A(j,i) is the transition probability from state
%       i to state j.
%   B = singleton potentials (biases).  B(j,t) is the singleton
%       or bias potential of state j at time t.  For an HMM,
%       B(j,t) is the likelihood of state j at time t (i.e. the 
%       probability of the observation at time t arising from state j).
%   op = an integer specifying the desired return value
%       -1: the mode (max-product)
%       0: marginals (sum-product) [default]
%       n: (integer >= 1) n samples from the posterior

% Author: David Ross, Oct. 2006
% $Id: loghmm.m,v 1.1 2006/10/31 20:49:20 dross Exp $

%-----------------------------------------------------------
% Check the input arguments
%-----------------------------------------------------------

error(nargchk(2,3,nargin));    

% by default, compute the marginals
if nargin < 3, op = 0; end
    
[J,T] = size(B);    

if any(size(A) ~= [J J])
    error('sizes of A and B do not match');
end

if numel(op) ~= 1
    error('op must be a scalar');
end

%-----------------------------------------------------------
% Find the mode (aka Max-Product or Viterbi)
%-----------------------------------------------------------
if op == -1

    mode = zeros(J,T);
    tau = zeros(J,T); % back pointers
    rho = zeros(J,T); % max values
    o = ones(1,J);

    % % in the probability domain
    % ERROR! this doesn't exactly equal the results in the log domain
    % Maybe the scaling is wrong.
    % rho(:,1) = B(:,1); % no prior
    % rho(:,1) = B(:,1) .* max(A/J, [], 2);
    % for tt = 2:T
    %     rho(:,tt-1) = rho(:,tt-1)/sum(rho(:,tt-1));
    %     [rho(:,tt), tau(:,tt)] = max(A' .* (rho(:,tt-1)*o), [], 1);
    %     rho(:,tt) = rho(:,tt) .* B(:,tt);
    %     if any(rho(:,tt) == 0)
    %         tt
    %     end
    % end
    % tau_old = tau;
    
    % in the log probability domain
    % (already there) A = log(A); B = log(B);
    rho = zeros(size(rho));
    rho(:,1) = B(:,1); % no prior
    rho(:,1) = B(:,1) + max(A + log(1/J), [], 2); % prior matches hmmviterbi
    for tt = 2:T
        [rho(:,tt), tau(:,tt)] = max(A' + rho(:,tt-1)*o, [], 1);
        rho(:,tt) = rho(:,tt) + B(:,tt);
    end

    % sum(tau(:) ~= tau_old(:))
    
    % backwards
    [junk mm] = max(rho(:,T));
    mode(mm,T) = 1;
    for tt = T-1:-1:1
        mm = tau(mm, tt+1);
        mode(mm,tt) = 1;
    end

    retval = mode;
    return; % FINISHED computing mode
end

%% If we're not doing Viterbi, shift from the log domain to the 
%% probability domain.
% perform and additive shift of A and B to prevent overflow
A = A - max(A(:)) + 1;
B = B - max(B(:)) + 1;
A = exp(A);
B = exp(B);


%-----------------------------------------------------------
% Forward recursion (alpha), required for marginals and
% for sampling.
%-----------------------------------------------------------

alpha = zeros(J,T);
scale(1) = sum(B(:,1));
alpha(:,1) = B(:,1) ./ scale(1);
for tt = 2:T
    alpha(:,tt) = A*alpha(:,tt-1) .* B(:,tt);
    scale(tt) = sum(alpha(:,tt));
    alpha(:,tt) = alpha(:,tt) ./ scale(tt);
end

% make sure there wasn't any underflow
if any(sum(abs(alpha),1) < eps)
    error('alpha goes to zero (rescaling is bad)');
end


%-----------------------------------------------------------
% Marginals
%-----------------------------------------------------------

if op == 0
    
    % backward
    beta = zeros(J,T);
    beta(:,T) = 1;
    for tt = T-1:-1:1
        beta(:,tt) = A'*(beta(:,tt+1) .* B(:,tt+1)) ./ scale(tt+1);
    end
    
    % make sure there wasn't any underflow
    if any(sum(abs(beta),1) < eps)
        error('beta goes to zero (rescaling is bad');
    end

    % marginals
    marginals = normalize(alpha .* beta, 1);

    % make sure they are non-negative and sum to (approximately) one
    if any(marginals(:) < 0)
        error('some marginals are negative!');
    elseif any((sum(abs(marginals),1)-1) > 100*eps)
        error('marginals do not sum to 1');
    end

    retval = marginals;
    if nargout > 1
        UUt = ((beta(:,2:end).*B(:,2:end))*alpha(:,1:end-1)') .* A;
    end
    return; % FINISHED computing marginals

%-----------------------------------------------------------
% Sample
%-----------------------------------------------------------

elseif op >= 1

    % % draw one sample
    % sample = logical(zeros(J,T));
    % ss = sample_discrete(alpha(:,T) ./ sum(alpha(:,T)),1);
    % sample(ss,T) = 1;
    % for tt = T-1:-1:1
    %     pp = alpha(:,tt) .* A(ss,:)';
    %     pp = pp ./ sum(pp);
    %     ss = sample_discrete(pp,1);
    %     sample(ss,tt) = 1;
    % end
    % retval = sample;
    % return; % FINISHED drawing a sample
    
    % draw many samples
    NS = double(op); % number of samples to draw
    o = ones(1,NS);
    %samples = repmat(logical(0), [T,J,NS]);
    samples = zeros(T,J,NS);    
    ss = sample_discrete(alpha(:,T), NS)';
    samples(T, ss+J*(0:NS-1)) = 1; % magic recipie for setting samples
    for tt = T-1:-1:1
        pp = (alpha(:,tt)*o) .* A(ss,:)';
        pp = normalize(pp,1);
        ss = sample_discrete2(pp,1);
        samples(tt, ss+J*(0:NS-1)) = 1;
    end
    retval = permute(samples, [2 1 3]);
    return; % FINISHED drawing samples

else
    error('unknown operation');
end

