function [mu,sigma,mu_pair,sigma_pair] = prob_X_given_UVY(Y,U,V,alpha,beta,varargin)
% [mu,sigma,mu_pair,sigma_pair] = prob_X_given_UVY(Y,U,V,alpha,beta, ...)
% [nu,prec,nu_pair,prec_pair] = prob_X_given_UVY(Y,U,V,alpha,beta, 'precision')
%
% Required Arguments:
%   Y = observation matrix (?*T)
%   U = matrix of binary 'switches' for dynamics features (J*(T-1))
%   V = matrix of binary 'switches' for observation features (K*T)
%   alpha = weights on dynamics features (D*D*J)
%   beta  = weights on observation features (D*D*K)
%
% Optional Argument:
%   'filter-only' = Only do the forward pass.  mu_pair and sigma_pair are
%                   not computed.
%   'precision' = Return the precision matrices and (@TODO what is it 
%      called) instead of the means and covariances.
%
% Return Values:
%   mu = the marginal means (D*T)
%   sigma = the marginal covariances (D*D*T)
%   mu_pair = the pairwise-marginal means (2D*(T-1))
%   sigma_pair = the pairwise-marginal covariances (2D*2D*(T-1))
%

% Author: David Ross, Oct. 2005.
% $Id: prob_X_given_UVY.m,v 1.6 2006/10/04 22:36:09 dross Exp $

U = max(U,1e-8); % @HACK to help with matrix conditioning
V = max(V,1e-8); % @HACK to help with matrix conditioning
% U = double(U); % @HACK
% V = double(V); % @HACK

%-------------------------------------------------------
% Check the input.
%-------------------------------------------------------

error(nargchk(5,7,nargin));

filter_only = 0;
moment_parameters = 1;
for ii = 1:length(varargin)
    if strcmp(varargin{ii}, 'filter-only')
        filter_only = 1;
    elseif strcmp(varargin{ii}, 'precision')
        moment_parameters = 0;
    else
        error(['unknown argument ' varargin{ii}]);
    end
end

% new version of Y
nu = cat(3,Y.nu);
nu = permute(nu, [1 3 2]);

% if beta is diagonal, we can invert it more easily
diag_beta = 1;
if any(any(any(beta .* (1-repmat(eye(size(beta,1)),[1 1 size(beta,3)])))))
    diag_beta = 0;
end

% @TODO add more checks of the input


%-------------------------------------------------------
% Initialization
%-------------------------------------------------------

% Should we try computing both inference updates and compare
% their numerical results?
check_other_method = 0;
tol = 1e-6;

D = size(alpha,1);
J = size(alpha,3);
% K = size(beta,3); % don't need it
T = size(Y,2);

% the mean and precision of the foward messages
msgf_nu = zeros(D,T);
msgf_prec = zeros(D,D,T);

% the corrected mean and precision of the forward pass
fwd_nu = zeros(D,T);
fwd_prec = zeros(D,D,T);

% the mean and precision of the messages from the observations
msgo_nu = zeros(D,T);
msgo_prec = zeros(D,D,T);

% the mean and precision of the backward messages
msgb_nu = zeros(D,T);
msgb_prec = zeros(D,D,T);

% the corrected mean and precision of the backward pass
bwd_nu = zeros(D,T);
bwd_prec = zeros(D,D,T);

% the pairwise marginal means and precisions
if nargout > 2
    prec_pair = zeros(2*D,2*D,T-1);
    nu_pair = zeros(2*D,T-1);
end

%% Form the mean and covariance of the dynamics term
% dyn_nu = zeros(2*D,1);
[Trans,dd] = T_matrices;
dd = reshape(dd, [D 1 J]);

alpha_trans = init_A_aT_TaT(alpha,Trans);
alpha_d = mult3(alpha,dd);
trans_alpha_d = mult3(permute(Trans,[2 1 3]), alpha_d);


%-------------------------------------------------------
% Compute the messages.
%-------------------------------------------------------

%% Compute the messages from the observations
for tt = 1:T
    % Fast way:
    msgo_prec(:,:,tt) = combine(beta,V(:,tt));
    msgo_nu(:,tt) = combine(mult3(beta,nu(:,tt,:)), V(:,tt));

    % % Slow way
    % for kk = 1:K
    %     msgo_prec(:,:,tt) = msgo_prec(:,:,tt) + beta(:,:,kk)*V(kk,tt);
    %     msgo_nu(:,tt) = msgo_nu(:,tt) + V(kk,tt)*beta(:,:,kk)*nu(:,tt,kk);
    % end
end

%% Initialize the first forward message
[mu0,var0] = get_mu0_var0;
msgf_prec(:,:,1) = inv(var0);
msgf_nu(:,1) = msgf_prec(:,:,1)*mu0;

fwd_prec(:,:,1) = msgf_prec(:,:,1) + msgo_prec(:,:,1);
fwd_nu(:,1) = msgf_nu(:,1) + msgo_nu(:,1);

%% Propagate the foward messages
for tt = 2:T
    [A,aT,TaT] = compute_A_aT_TaT(alpha_trans,U(:,tt-1));
    Tad = combine(trans_alpha_d, U(:,tt-1));
    ad = combine(alpha_d, U(:,tt-1));
    
    % predict
    Ai = inv(A);
    Ai_aT = Ai*aT;
    msgf_prec(:,:,tt) = inv(Ai + Ai_aT/(TaT + fwd_prec(:,:,tt-1) - aT'*Ai_aT)*Ai_aT');
    msgf_nu(:,tt) = ad + aT/(TaT+fwd_prec(:,:,tt-1))*(fwd_nu(:,tt-1) - Tad);

    if check_other_method 
        p = msgf_prec(:,:,tt);
        msgf_prec(:,:,tt) = A - aT/(TaT+fwd_prec(:,:,tt-1))*aT';
        if sse(p - msgf_prec(:,:,tt)) > tol
            warning('they don''t match');
        end
    end
    
    % correct
    fwd_prec(:,:,tt) = msgf_prec(:,:,tt) + msgo_prec(:,:,tt);
    fwd_nu(:,tt) = msgf_nu(:,tt) + msgo_nu(:,tt);
end

%% Return now if we're only doing filtering
if filter_only
    if moment_parameters
        [mu,sigma] = canonical_to_moment(fwd_nu,fwd_prec);
    else
        mu = fwd_nu;
        sigma = fwd_prec;
    end
    return
end

%% Initialize the backward messages
bwd_prec(:,:,T) = msgb_prec(:,:,T) + msgo_prec(:,:,T);
bwd_nu(:,T) = msgb_nu(:,T) + msgo_nu(:,T);

%% Propagate the backward messages
for tt = T-1:-1:1
    [A,aT,TaT] = compute_A_aT_TaT(alpha_trans,U(:,tt));
    Tad = combine(trans_alpha_d, U(:,tt));
    ad = combine(alpha_d, U(:,tt));

    % predict (gpp requires that bwd be padded)
    TaTi = inv(TaT);
    aT_TaTi = aT*TaTi;
    msgb_prec(:,:,tt) = inv(TaTi + ...
        aT_TaTi'/(A + bwd_prec(:,:,tt+1) - aT_TaTi*aT')*aT_TaTi);
    msgb_nu(:,tt) = -Tad + aT'/(A+bwd_prec(:,:,tt+1))*(bwd_nu(:,tt+1) + ad);
    
    if check_other_method 
        p = msgb_prec(:,:,tt);
        msgb_prec(:,:,tt) = TaT - aT'/(A + bwd_prec(:,:,tt+1))*aT;
        if sse(p - msgb_prec(:,:,tt)) > tol
            warning('they don''t match');
        end
    end
    
    % correct
    bwd_prec(:,:,tt) = msgb_prec(:,:,tt) + msgo_prec(:,:,tt);
    bwd_nu(:,tt) = msgb_nu(:,tt) + msgo_nu(:,tt);
    
    % collect pairwise precision and nu
    if nargout > 2
        prec_pair(:,:,tt) = [fwd_prec(:,:,tt)+TaT, -aT';
            -aT, bwd_prec(:,:,tt+1)+A];
        nu_pair(:,tt) = [fwd_nu(:,tt)-Tad; ad + bwd_nu(:,tt+1)];
    end
end

%-------------------------------------------------------
% Combine messages to compute the marginals.
%-------------------------------------------------------

%% Compute marginals and
%% Convert precisions to covariances
prec_marg = fwd_prec + msgb_prec;
nu_marg = fwd_nu + msgb_nu;

if moment_parameters
    [mu,sigma] = canonical_to_moment(nu_marg,prec_marg);
    if nargout > 2
        [mu_pair,sigma_pair] = canonical_to_moment(nu_pair,prec_pair);
    end
else
    mu = nu_marg;
    sigma = prec_marg;
    if nargout > 2
        mu_pair = nu_pair;
        sigma_pair = prec_pair;
    end
end
