function [X] = sample_x_unigauss(Y,U,V,alpha,beta,N)
% [X] = sample_x_unigauss(Y,U,V,alpha,beta)
% [X] = sample_x_unigauss(Y,U,V,alpha,beta,N)
% Draw N samples from P(X|Y,U,V,alpha,beta)
%
% Arguments:
%   Y = an observation sequence given as column vectors
%   alpha, beta = the weights
%   N = # of samples to draw (default N=1)
%
% Return Value:
%   X = a D*T*N matrix of N samples from the joint P(X|Y,U,V,alpha,beta).

% Author: David Ross, Oct. 2005
% $Id: sample_x_unigauss.m,v 1.3 2006/07/21 15:54:37 dross Exp $

U = max(U,1e-8); % @HACK to help with matrix conditioning
V = max(V,1e-8); % @HACK to help with matrix conditioning

%----------------------------------------------------------
% Check the input
%----------------------------------------------------------

error(nargchk(5,6,nargin));
if nargin < 6, N=1; end


%----------------------------------------------------------
% Do the sampling
%----------------------------------------------------------

% Version 1 == correct sampling using the pairwise marginals
% Version 2 == potentially faster sampling which only requires
%              forward filtering (@TODO unfinished)
% Version 3 == just sample from the marginals (incorrect but fast)
version = 1;

%---------------------
% Sampling version 1
%---------------------
if version == 1
    % % old way
    % [mu,sigma,mu_pair,sigma_pair] = prob_X_given_UVY(Y,U,V,alpha,beta);
    % [D,T] = size(nu);
    % J = size(alpha,3);
    % X = zeros(D,N,T);
    % o = ones(1,N);
    % X(:,:,T) = gaussrnd(mu(:,T), sigma(:,:,T), N);
    % for tt = T-1:-1:1
    %     ccv = sigma_pair(1:D,D+1:end,tt);
    %     C = sigma(:,:,tt) - ccv*inv(sigma(:,:,tt+1))*ccv';
    %     % alternatively, look only at sigma_pair
    %     % ul = sigma_pair(1:D,1:D,tt);
    %     % lr = sigma_pair(D+1:end,D+1:end,tt);
    %     % C = ul - ccv*inv(lr)*ccv';
    %     m = mu(:,tt)*o + ...
    %         ccv*inv(sigma(:,:,tt+1))*(X(:,:,tt+1) - mu(:,tt+1)*o);
    %     X(:,:,tt) = chol(C)'*randn(D,N) + m;
    % end
         
    [nu,prec,nu_pair,prec_pair] = prob_X_given_UVY(Y,U,V,alpha,beta,'precision');    
    [D,T] = size(nu);
    J = size(alpha,3);
    X = zeros(D,N,T);
    o = ones(1,N);
    X(:,:,T) = gaussrnd(prec(:,:,T)\nu(:,T), inv(prec(:,:,T)), N);
    for tt = T-1:-1:1
        C = inv(prec_pair(1:D,1:D,tt));
        m = C*(nu_pair(1:D,tt)*o - prec_pair(1:D,D+1:end,tt)*X(:,:,tt+1));
        X(:,:,tt) = chol(C)'*randn(D,N) + m;
    end
         
%---------------------
% Sampling version 2
%---------------------
elseif version == 2
    % do forward message passing, to compute P(x_t|y_{1:t}) for all t.
    [mu,sigma] = prob_X_given_UVY(Y,U,V,alpha,beta,'filter-only');

    [D,T] = size(mu);
    J = size(alpha,3);
    X = zeros(D,N,T);
    o = ones(1,N);
    Trans = T_matrices;

    % start sampling at the end of the sequence
    X(:,:,T) = gaussrnd(mu(:,T), sigma(:,:,T), N);

    alpha_trans = init_A_aT_TaT(alpha,Trans);

    for tt = T-1:-1:1
        [A,aT,TaT] =  compute_A_aT_TaT(alpha_trans, U(:,tt));

        prec1 = inv(sigma(:,:,tt));
        prec2 = TaT - aT'*inv(A)*aT;
        C = inv(prec1 + prec2);
        this_mu = C *(prec1*mu(:,tt)*o + prec2*(-aT'*(A\X(:,:,tt+1))));
        X(:,:,tt) = chol(C)'*randn(D,N) + this_mu;
    end

    error(['@TODO This hasn''t been fully implemented' ...
           ' and ignores the dd from T_matrices']);

%---------------------
% Sampling version 3
%---------------------
elseif version == 3

    % do inference to compute the marginals
    [mu,sigma] = prob_X_given_UVY(Y,U,V,alpha,beta);
    [D,T] = size(mu);
    X = zeros(D,N,T);
    o = ones(1,N);

    % sample from the marginals
    for tt = 1:T
        X(:,:,tt) = chol(sigma(:,:,tt))*randn(D,N) + mu(:,tt)*o;
    end

else
    error('invalid version');
end

X = permute(X,[1 3 2]);
